/*
 * DEMOORDM.CPP - Demo code on how to use the usbdmx.dll to talk to
 * RDM devices.
 * Link you code against usbdmx.lib.
 *
 * This code follows the "ANSI E1.20 - 2010 Entertainment TechnologyRDM
 * Remote Device Management Over DMX512 Networks".
 *
 * The usbdmx.dll provide two functions to support RDM:
 * - usbdmx_rdmdiscovery() for discovery (DISC_UNIQUE_BRANCH) messages which
 *   use a different protocol for the reply and requires more response 
     processing.
 * - usbdmx_rdm() for all other RDM messages.
 * Both functions transmit START Code, Sub START Code and Message Length
 * followed by the user supplied buffer completed by the 16bit Checksum. So the
 * user buffer should not contain this data. The return value is either a
 * special return value (<0) indicating errors or a positive value indicating
 * the number of bytes of successfully received and processed data. Both
 * functions return processed data. usbdmx_rdmdiscovery() returns a list of
 * UIDs an usbdmx_rdm() the response message content without START Code, Sub
 * START Code, Message Length and Checksum.
 * More details will be described in the code.
 *
 * This file is provided as is to allow an easy start with the
 * usbdmx driver and dll.
 *
 * In case of trouble please contact driver@lighting-solutions.de or
 * call +49/40/600877-51.
 */

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdio.h>

/* the rdm.h include file provides most RDM related defines and can be found at
 * http://www.rdmprotocol.org */
#include "rdm.h"

 /* the usbdmx include file with all function definitions */
#include "../../usbdmx.dll/usbdmx.h"

/* define a type that describes the content of RDM messages */
typedef struct {
	/* the len-field is used in this code to handle the RDM message length
	 * together with the data. */
	UCHAR	len;

	/* this is the RDM message content thats send/received to the usbdmx.dll
	 * !! the usbdmx.dll adds START Code, Sub START Code, Message Length and
	 *    16bit Checksum. So this values are omitted here. */
	struct {
		/* this fields are added by the driver
		  UCHAR	StartCode;
		  UCHAR	SubStartCode;
		  UCHAR	MessageLength;
		 */
		UCHAR	DestinationUID[6];	/* MSB first */
		UCHAR	SourceUID[6];
		UCHAR	TransactionNumber;
		union {
			UCHAR	PortId;
			UCHAR	ResponseType;
		};
		UCHAR	MessageCount;
		UCHAR	SubDevice[2];		/* MSB first */
		UCHAR	CommandClass;
		UCHAR	ParameterID[2];
		UCHAR	ParameterDataLength;
		UCHAR   data[255 - 24];		/* maximum size is 255 incl. start codes and length field */
	} packet;					    /* the actual RDM packet */
} RdmPacket_T;

/* define a type to describe UIDs in this code */
typedef ULONG64	Uid_T;

/* define the source UID the code uses */
#define SOURCE_UID	0x010000000003ULL

/* define structure to store information about discovered devices */
typedef struct {
	Uid_T  Uid;			/* this is the Uid of the devices returned by the discovery process */
	USHORT control;		/* control field returned by mute/unmute commands */

	/* the following information is retrieved as part of this demonstration */
	struct {
		USHORT RDMProtocolVersion;
		USHORT DeviceModelId;
		USHORT ProductCategory;
		UINT   SoftwareVersionId;
		USHORT DMX512Footprint;
		UCHAR  DMX512CurrentPersonality, DMX512NumPersonalities;
		USHORT DMX512StartAddress;
		USHORT SubDeviceCount;
		UCHAR  SensorCount;
	} DeviceInfo;

	/* software version label - standard specifies max. length of 32 characters */
	char SoftwareVersionLabel[33];
} RdmDevice_T;

/* this is the Table-of-Devices discovered */
static RdmDevice_T ToD[32];
/* and the amount of device in this table */
static unsigned int TodDevices;

/* split a 16-bit value into two 8bit values to initialize an array */
#define U16(a)	(unsigned char)(((a) & 0xff00) >> 8), (unsigned char)((a) & 0x00ff)

/* split a uid into six 8bit values to initialize an array */
#define UID(a)	(unsigned char)(((a) >> 40) & 0xff), \
				(unsigned char)(((a) >> 32) & 0xff), \
				(unsigned char)(((a) >> 24) & 0xff), \
				(unsigned char)(((a) >> 16) & 0xff), \
				(unsigned char)(((a) >>  8) & 0xff), \
				(unsigned char)(((a) >>  0) & 0xff) 

/* get a uid from pBuf - in RDM format, Big-Endian, MSB first */
static Uid_T GetUid(const void *pBuf)
{
	const UCHAR* p = (const UCHAR *)pBuf;
	Uid_T Uid;

	Uid = 0UL;
	Uid  = (Uid_T)*p++ << 40;
	Uid |= (Uid_T)*p++ << 32;
	Uid |= (Uid_T)*p++ << 24;
	Uid |= (Uid_T)*p++ << 16;
	Uid |= (Uid_T)*p++ <<  8;
	Uid |= (Uid_T)*p++ <<  0;

	return Uid;
}

/* get a 16bit unsigned from pBuf - in RDM format, Big-Endian, MSB first */
static USHORT GetU16(const void* pBuf)
{
	const UCHAR* p = (const UCHAR*)pBuf;
	return (p[0] << 8) | p[1];
}

/* get a 32bit unsigned from pBuf - in RDM format, Big-Endian, MSB first */
static UINT GetU32(const void* pBuf)
{
	const UCHAR* p = (const UCHAR*)pBuf;
	return (p[0] << 24) | (p[1] << 16) | (p[2] << 8) | p[3];
}

/* convert NACK reasons to human readable strings */
static const char* NackReasonStr(USHORT reason)
{
	switch (reason)
	{
	case E120_NR_UNKNOWN_PID:				return "Unknown PID";
	case E120_NR_FORMAT_ERROR:				return "Format Error";
	case E120_NR_HARDWARE_FAULT:			return "Hardware Fault";
	case E120_NR_UNSUPPORTED_COMMAND_CLASS:	return "Unsupported Command Class";
	case E120_NR_DATA_OUT_OF_RANGE:			return "Data out of range";
	case E120_NR_SUB_DEVICE_OUT_OF_RANGE:	return "Sub-Device out of range";
	default:								return "unknown";
	}
}

/*
 * AnalyseRespone - compare transmitted and received RDM message and flag any unexpected value
 */
static int AnalyseResponse(RdmPacket_T* ptx, RdmPacket_T* prx)
{
	if (GetUid(ptx->packet.DestinationUID) != GetUid(prx->packet.SourceUID))	/* destination Uid */
	{
		printf("ERROR: Source-UID\n");
		return -100;
	}
	if (GetUid(ptx->packet.SourceUID) != GetUid(prx->packet.DestinationUID))	/* source Uid */
	{
		printf("ERROR: Destination-UID\n");
		return -101;
	}
	if (ptx->packet.TransactionNumber != prx->packet.TransactionNumber)			/* transaction number */
	{
		printf("ERROR: Transaction Number\n");
		return -102;
	}
	if (ptx->packet.CommandClass + 1 != prx->packet.CommandClass)				/* command class */
	{
		printf("ERROR: Command Class\n");
		return -103;
	}
	if (GetU16(ptx->packet.ParameterID) != GetU16(prx->packet.ParameterID))		/* parameter ID */
	{
		printf("ERROR: ParameterID\n");
		return -104;
	}

	return 0;
}

/*
 * RdmGet - send a rdm get message
 */
static int RdmGet(HANDLE h, USHORT ParameterID, Uid_T uid, RdmPacket_T *rx)
{
	/* buffer for the rdm message to be send to the device */
	RdmPacket_T tx = {
		21,
		{
			UID(uid),			/* destination Uid - will be replaced later */
			UID(SOURCE_UID),	/* source Uid */
			0x00,				/* transaction number */ 
			0x01,				/* port ID */
			0x00,				/* message count */
			U16(1),				/* Sub-Device */
			E120_GET_COMMAND,	/* command class - GET */
			U16(ParameterID),	/* parameter ID */
			0x00				/* parameter data length */
		}
	};
	int ret;

	/* send message */
	ret = usbdmx_rdm(h, 0, tx.len, &tx.packet, sizeof(rx->packet), &rx->packet);
	/* analyse return value - details can be found in usbdmx.dll USBDMX_RDM_<*> */
	if (ret < 0)
	{	/* < 0 -> transmission/reception error */
		return ret;
	}
	/* no error -> ret is the number of bytes received */
	rx->len = ret;

	/* analyze results - compare transmitted and received message */
	ret = AnalyseResponse(&tx, rx);
	if (ret)
		return ret;

	/* if the response is not an acknowledge, try to decode it for debugging */
	if (rx->packet.ResponseType != E120_RESPONSE_TYPE_ACK)
	{
		if (rx->packet.ResponseType == E120_RESPONSE_TYPE_NACK_REASON)
		{
			if (rx->packet.ParameterDataLength != 2)
			{
				printf("ERROR: get(): NACK: ParameterDataLength = %u\n", rx->packet.ParameterDataLength);
				return -110;
			}
			USHORT reason = GetU16(rx->packet.data);

			printf("WARN: RdmGet: ParameterID 0x%04x: NACK 0x%04x (%s)\n", ParameterID, reason, NackReasonStr(reason));
			return -111;
		}

		printf("ERROR: RdmGet: ParameterID 0x%04x: ResponseType 0x%02x\n", ParameterID, rx->packet.ResponseType);
	}

	return 0;
}

/*
 * RdmSet - send a rdm set message
 */
static int RdmSet(HANDLE h, USHORT ParameterID, Uid_T uid, UCHAR txdatalen, const void *txdata, RdmPacket_T *rx)
{
	RdmPacket_T tx = {
		21,
		{
			UID(uid),			/* destination Uid */
			UID(SOURCE_UID),	/* source Uid */
			0x00,				/* transaction number */
			0x01,				/* port ID - 1..255 */
			0x00,				/* message count */
			U16(1),				/* Sub-Device */
			E120_SET_COMMAND,	/* command class - SET */
			U16(ParameterID),	/* parameter ID */
			0x00				/* parameter data length - currently 0, the data to be send will be added next */
		}
	};
	int ret;

	/* limit data length */
	if (txdatalen > sizeof(tx.packet.data))
		txdatalen = sizeof(tx.packet.data);

	/* add data to message */
	tx.packet.ParameterDataLength = txdatalen;
	memcpy(&tx.packet.data, txdata, txdatalen);
	/* adjust frame length */
	tx.len += txdatalen;

	/* send message */
	ret = usbdmx_rdm(h, 0, tx.len, &tx.packet, sizeof(rx->packet), &rx->packet);
	/* analyze return value - details can be found in usbdmx.dll USBDMX_RDM_<*> */
	if (ret < 0)
	{	/* < 0 -> transmission/reception error */
		return ret;
	}
	/* no error -> ret is the number of bytes received */
	rx->len = ret;

	/* analyze results - compare transmitted and received message */
	ret = AnalyseResponse(&tx, rx);
	if (ret)
		return ret;

	/* if the response is not an acknowledge, try to decode it for debugging */
	if (rx->packet.ResponseType != E120_RESPONSE_TYPE_ACK)
	{
		if (rx->packet.ResponseType == E120_RESPONSE_TYPE_NACK_REASON)
		{
			if (rx->packet.ParameterDataLength != 2)
			{
				printf("ERROR: set(): NACK: ParameterDataLength = %u\n", rx->packet.ParameterDataLength);
				return -110;
			}
			USHORT reason = GetU16(rx->packet.data);

			printf("WARN: RdmSet: ParameterID 0x%04x: NACK 0x%04x (%s)\n", ParameterID, reason, NackReasonStr(reason));
			return -111;
		}

		printf("ERROR: RdmSet: ParameterID 0x%04x: ResponseType 0x%02x\n", ParameterID, rx->packet.ResponseType);
	}

	return 0;
}

/*
 * send a mute/unmute command. This is a dedicated functions as SubDevice has to be 0
 */
static int muteUnmute(HANDLE h, Uid_T uid, unsigned int cmdMute)
{
	const USHORT ParameterID = cmdMute ? E120_DISC_MUTE : E120_DISC_UN_MUTE;
	RdmPacket_T tx = {
		21,
		{
			UID(uid),				/* destination Uid */
			UID(SOURCE_UID),		/* source Uid */
			0x00,					/* transaction number */
			0x01,					/* port ID - 1..255 */
			0x00,					/* message count */
			U16(0x00), 				/* Sub-Device */
			E120_DISCOVERY_COMMAND,	/* command class */
			U16(ParameterID),		/* parameter ID */
			0x00					/* parameter data length */
		}
	};
	RdmPacket_T rx;
	int ret;

	/* send message */
	if (uid != E120_BROADCAST_ALL_DEVICES_ID)
		ret = usbdmx_rdm(h, 0, tx.len, &tx.packet, sizeof(rx.packet), &rx.packet);
	else
		ret = usbdmx_rdm(h, 0, tx.len, &tx.packet, 0, 0);
	/* check the return value: < 0 -> transmit/receive error */
	if (ret < 0)
		return ret;

	/* the message was send to a specific devices, analyze the response */
	if (uid != E120_BROADCAST_ALL_DEVICES_ID)
	{
		/* the message was send successful -> ret is the number of bytes received */
		rx.len = ret;

		/* analyze the response */
		ret = AnalyseResponse(&tx, &rx);
		if (ret)
			return ret;

		/* mute/unmute message can contains 2 or more data bytes, decode the control field */
		if (rx.packet.ParameterDataLength >= 2)
			/* control field found -> decode and return it */
			ret = GetU16(rx.packet.data);
	}

	return ret;
}

/*
 * Send a unique branch message with UIDs set to min and max
 */
static int SendUniqueBranch(HANDLE h, Uid_T UidMin, Uid_T UidMax, UCHAR rxsize, void *rxbuffer)
{
	RdmPacket_T tx = {							/* this is the message that will be send out */
		33,
		{
			UID(E120_BROADCAST_ALL_DEVICES_ID),	/* Destination UID is broadcast */
			UID(SOURCE_UID),					/* Source Uid */
			0x00,								/* transaction number */
			0x01,								/* port ID - 1..255 */
			0x00,								/* message count */
			U16(0),								/* Sub-Device */
			E120_DISCOVERY_COMMAND,				/* command class */
			U16(E120_DISC_UNIQUE_BRANCH), 		/* parameter ID */
			0x0C,								/* parameter data length - two UIDs */
			{
				UID(UidMin),					/* lower boundary UID */
				UID(UidMax)						/* upper boundary UID */
			}
		}
	};

	/* send frame using RDM discovery function */
	return usbdmx_rdmdiscovery(h, 0, tx.len, &tx.packet, rxsize, rxbuffer);
}

/*
 * Discovery - full binary discovery search. This function is called recursively
 */
static int DiscoveryDo(HANDLE h, Uid_T UidMin, Uid_T UidMax)
{
	UCHAR  devices[18];		/* response buffer - support up to 3 devices with one discovery */
	UCHAR *p;				/* pointer to loop through the response buffer */
	int    ret;				
	int    NumDevices = 0;	/* number of devices found */
	int	   NewDevices;
	Uid_T  uid;				/* discovered Uid */
	unsigned int tries;

	NumDevices = 0;
	for (tries = 3; tries > 0; tries--)
	{
		/* send unique branch with given Uid range */
		ret = SendUniqueBranch(h, UidMin, UidMax, sizeof(devices), devices);
		/* if the response could be a collision, split the range into two branches and retry */
		if (ret == USBDMX_RDM_COLLISION)
		{
			/* calculate the Uid in the middle between Min and Max */
			Uid_T UidMid = (UidMax + UidMin) / 2;

			/* search each branch */
			if (UidMin <= UidMid)
			{
				ret = DiscoveryDo(h, UidMin, UidMid);
				if (ret < 0)
					return ret;
				if (ret > 0)
					NumDevices += ret;
			}
			if (UidMid < UidMax)
			{
				ret = DiscoveryDo(h, UidMid + 1, UidMax);
				if (ret < 0)
					return ret;
				if (ret > 0)
					NumDevices += ret;
			}
		}
		/* other errors? */
		else if (ret < 0)
		{
			/* transmission/reception failed */
			return ret;
		}
		else
		{
			/* no errors -> device('s) found */
			p = devices;			/* the device buffer contains a list of Uids or is empty */
			NewDevices = ret / 6;	/* the return value is the number of bytes written to devices buffer, 6 bytes per Uid, so NewDevices is the number of Uids in devices */
			/* loop to process all Uids in devices buffer */ 
			while (NewDevices > 0)
			{
				uid = GetUid(p);	/* extract the next Uid from devices buffer */
				p += 6;				/* forward buffer pointer p to next Uid */
				NewDevices -= 1;

				/* test if the device with uid exists by muting it */
				ret = muteUnmute(h, uid, 1);	/* this function returns the control field from mute response, if given */
				if (ret >= 0)
				{
					printf("Found device %04x:%08x\n", (unsigned int)(uid >> 32), (unsigned int)(uid & 0xffffffffU));

					/* if there is space, add the device to the Table of Devices */
					if (TodDevices < ARRAYSIZE(ToD))
					{
						ToD[TodDevices].Uid = uid;		/* remember the device we just found */
						ToD[TodDevices].control = ret;	/* remember the control field */
						TodDevices++;
					}
					NumDevices++;
				}
			}
		}
	}
	return NumDevices;
}

/*
 * DISCOVERY - do full RDM discovery using Binary Search Tree
 * This code does not fully follow the RDM recommendation in that
 * it does not do retries and is not tested agains common device
 * implementation errors. See http://www.lightingandsoundamerica.com/mailing/PLASAProtocol/PSpring12_RDMCornerCases.pdf
 * for "Corner Cases with RDM Discovery"
 */
static BOOL discovery(HANDLE h)
{
	int devices;	/* return value from binary tree search */

	/* reset out global table of Devices */
	TodDevices = 0;

	/* send unmute to all devices */
	muteUnmute(h, E120_BROADCAST_ALL_DEVICES_ID, 0);

	/* start discovery */
	devices = DiscoveryDo(h, (Uid_T)0, E120_BROADCAST_ALL_DEVICES_ID);
	if (devices < 0)
	{
		printf("ERROR: discovery failed with error %d\n", devices);
		return devices;
	}
	return devices;
}

/* send a GET DEVICE_INFO */
static int getDeviceInfo(HANDLE h, RdmDevice_T * pdev)
{
	RdmPacket_T rx;
	int ret;

	/* send the GET DEVICE_INFO message */
	ret = RdmGet(h, E120_DEVICE_INFO, pdev->Uid, &rx);

	/* if the transaction was successful, extract and remember the information we received */
	if (ret == 0)
	{
		/* check if the length is as expected */
		if (rx.packet.ParameterDataLength != 19)
		{
			printf("ERROR: getDeviceInfo(): ParameterDataLength error: got %u, expected %u\n", rx.packet.ParameterDataLength, 19);
			ret = -100;
		}
		else
		{
			/* this structure describes the structure of the received Device-Info message */
			struct DeviceInfo_S {
				UCHAR RDMProtocolVersion[2];
				UCHAR DeviceModelID[2];
				UCHAR ProductCategory[2];
				UCHAR SoftwareVersionID[4];
				UCHAR DMX512Footprint[2];
				UCHAR DMX512Personality[2];
				UCHAR DMX512StartAddress[2];
				UCHAR SubDeviceCount[2];
				UCHAR SensorCount;
			} const *DeviceInfo = (struct DeviceInfo_S *)rx.packet.data;	/* the data is in the data part of the received message */

			/* extract the data - use special get-functions to handle the byte order: RDM uses big-endian (MSB first) */
			pdev->DeviceInfo.RDMProtocolVersion = GetU16(DeviceInfo->RDMProtocolVersion);
			pdev->DeviceInfo.DeviceModelId = GetU16(DeviceInfo->DeviceModelID);
			pdev->DeviceInfo.ProductCategory = GetU16(DeviceInfo->ProductCategory);
			pdev->DeviceInfo.SoftwareVersionId = GetU32(DeviceInfo->SoftwareVersionID);
			pdev->DeviceInfo.DMX512Footprint = GetU16(DeviceInfo->DMX512Footprint);
			pdev->DeviceInfo.DMX512CurrentPersonality = DeviceInfo->DMX512Personality[0];
			pdev->DeviceInfo.DMX512NumPersonalities = DeviceInfo->DMX512Personality[1];
			pdev->DeviceInfo.DMX512StartAddress = GetU16(DeviceInfo->DMX512StartAddress);
			pdev->DeviceInfo.SubDeviceCount = GetU16(DeviceInfo->SubDeviceCount);
			pdev->DeviceInfo.SensorCount = DeviceInfo->SensorCount;
		}
	}

	return ret;
}

/* send a GET SOFTWARE_VERSION_LABEL */
static int getSoftwareVersionLabel(HANDLE h, RdmDevice_T* pdev)
{
	RdmPacket_T rx;
	int ret;
	unsigned int len;

	/* send the GET SOFTWARE_VERSION_LABEL message */
	ret = RdmGet(h, E120_SOFTWARE_VERSION_LABEL, pdev->Uid, &rx);

	/* if the transaction was successful, extract and remember the information we received */
	if (ret == 0)
	{
		/* check if the length is as expected */
		len = rx.packet.ParameterDataLength;
		if (len >= sizeof(pdev->SoftwareVersionLabel))
		{
			len = sizeof(pdev->SoftwareVersionLabel) - 1;
			printf("WARNING: getSoftwareVersionLabel(): ParameterDataLength error: got %u, expected < %u\n", rx.packet.ParameterDataLength, sizeof(pdev->SoftwareVersionLabel));
		}

		/* copy/remember software version label */
		memcpy(pdev->SoftwareVersionLabel, (const char *)rx.packet.data, len);
		pdev->SoftwareVersionLabel[len] = '\0';
	}

	return ret;
}

/* send a set DMX_START_ADDRESS */
static int setDmxStartAddress(HANDLE h, Uid_T uid, unsigned short address)
{
	RdmPacket_T rx;
	/* package the new start address into the RDM byte-order: Big-Endian (MSB first) */
	unsigned char buf[2] = { U16(address) };

	/* send the SET DMX_START_ADDRESS message */
	return RdmSet(h, E120_DMX_START_ADDRESS, uid, sizeof(buf), buf, &rx);
}

/* send a GET DMX_START_ADDRESS */
static int getDmxStartAddress(HANDLE h, Uid_T uid, PUSHORT padr)
{
	RdmPacket_T rx;
	int ret;

	/* send the GET GET_START_ADDRESS message */
	ret = RdmGet(h, E120_DMX_START_ADDRESS, uid, &rx);

	/* if the transaction was successful, extract and remember the information we received */
	if (ret == 0)
	{
		/* check if the length is as expected */
		if (rx.packet.ParameterDataLength != 2)
		{
			printf("getDmxStartAddress: ParameterDataLength != 2: %u.\n", rx.packet.ParameterDataLength);
			ret = -200;
		}

		if (padr)
			*padr = GetU16(rx.packet.data);
	}

	return ret;
}

/* send a SET IDENTIFY_DEVICE */
static int setIdentifyDevice(HANDLE h, Uid_T uid, unsigned char state)
{
	RdmPacket_T rx;
	return RdmSet(h, E120_IDENTIFY_DEVICE, uid, sizeof(state), &state, &rx);
}

/* send a get IDENTIFY_DEVICE */
static int getIdentifyDevice(HANDLE h, Uid_T uid, PUCHAR pstate)
{
	RdmPacket_T rx;
	int ret;

	/* send the GET IDENTIFY_DEVICE message */
	ret = RdmGet(h, E120_IDENTIFY_DEVICE, uid, &rx);

	/* if the transaction was successful, extract and remember the information we received */
	if (ret == 0)
	{
		/* check if the length is as expected */
		if (rx.packet.ParameterDataLength != 1)
		{
			printf("getIdentifyDevice: ParameterDataLength != 1: %u.\n", rx.packet.ParameterDataLength);
			ret = -200;
		}

		if (pstate)
			*pstate = rx.packet.data[0];
	}

	return ret;
}

/*
 * test/read a device with given uid
 */
static int testDevice(HANDLE h, RdmDevice_T *pdev)
{
	int ret;
	USHORT DmxStartAddressOld, DmxStartAddressSet, DmxStartAddressGet;
	UCHAR  IdentifyDeviceSet, IdentifyDeviceGet;

	printf("\ndevice %04x:%08x:\n", (unsigned int)(pdev->Uid >> 32), (unsigned int)(pdev->Uid & 0xffffffffUL));
	printf("  ControlField: 0x%04x\n", pdev->control);

	/* get Device information */
	ret = getDeviceInfo(h, pdev);
	if (ret)
		printf("error: getDeviceInfo failed with error %d.\n", ret);
	else
	{
		printf("  DeviceInfo:\n");
		printf("    RDMProtocolVersion: %u.%u\n", pdev->DeviceInfo.RDMProtocolVersion >> 8, pdev->DeviceInfo.RDMProtocolVersion & 0xff);
		printf("    DeviceModelId: %lu\n", pdev->DeviceInfo.DeviceModelId);
		printf("    ProductCategory: 0x%04x\n", pdev->DeviceInfo.ProductCategory);
		printf("    SoftwareVersionId: %lu\n", pdev->DeviceInfo.SoftwareVersionId);
		printf("    DMX512Footprint: %lu\n", pdev->DeviceInfo.DMX512Footprint);
		printf("    DMX512CurrentPersonality: %lu\n", pdev->DeviceInfo.DMX512CurrentPersonality);
		printf("    DMX512NumPersonalities: %lu\n", pdev->DeviceInfo.DMX512NumPersonalities);
		printf("    DMX512StartAddress: %lu\n", pdev->DeviceInfo.DMX512StartAddress);
		printf("    SubDeviceCount: %lu\n", pdev->DeviceInfo.SubDeviceCount);
		printf("    SensorCount: %lu\n", pdev->DeviceInfo.SensorCount);
	}

	/* get software version label */
	ret = getSoftwareVersionLabel(h, pdev);
	if (ret)
		printf("error: getSoftwareVersionLabel failed with error %d.\n", ret);
	else
	{
		printf("  SoftwareVersionLabel: %s\n", pdev->SoftwareVersionLabel);
	}

	/* test if the device consumes and DMX512 data - if so, has to support the GET/SET DMX_START_ADDRESS messages */
	if (pdev->DeviceInfo.DMX512Footprint > 0)
	{
		/* read back the current dmx512 start address */
		ret = getDmxStartAddress(h, pdev->Uid, &DmxStartAddressOld);
		if (ret)
			printf("error: getStartAddress failed with error %d.\n", ret);

		/* change the dmx512 start address to 1 - for demonstration purpose only */
		DmxStartAddressSet = 1;
		ret = setDmxStartAddress(h, pdev->Uid, DmxStartAddressSet);
		if (ret)
			printf("error: setStartAddress failed with error %d.\n", ret);

		/* read back dmx512 start address */
		ret = getDmxStartAddress(h, pdev->Uid, &DmxStartAddressGet);
		if (ret)
			printf("error: getStartAddress failed with error %d.\n", ret);

		/* compare set/get dmx512 start address */
		else if (DmxStartAddressGet != pdev->DeviceInfo.DMX512StartAddress)
			printf("error: getStartAddress returned wrong address. expected %u, got %u.\n", pdev->DeviceInfo.DMX512StartAddress, DmxStartAddressGet);

		/* restore old start address */
		ret = setDmxStartAddress(h, pdev->Uid, DmxStartAddressOld);
		if (ret)
			printf("error: setStartAddress failed with error %d.\n", ret);
	}

	// test IDENTIFY_DEVICE command
	IdentifyDeviceSet = 1;
	ret = setIdentifyDevice(h, pdev->Uid, IdentifyDeviceSet);
	if (ret)
		printf("error: setIdentifyDevice failed with error %d.\n", ret);

	ret = getIdentifyDevice(h, pdev->Uid, &IdentifyDeviceGet);
	if (ret)
		printf("error: getIdentifyDevice failed with error %d.\n", ret);

	// compare set/get device identification state
	else if (IdentifyDeviceGet != IdentifyDeviceSet)
		printf("error: getIdentifyDevice returned wrong state. expected %u, got %u.\n", IdentifyDeviceSet, IdentifyDeviceGet);

	return ret;
}

int main(int argc, char* argv[])
{
	USHORT device;			/* device number to open */
	HANDLE h;				/* handle to one interface */
	BOOL   ret;				/* return value */
	WCHAR  product[32];		/* device product string */
	USHORT version;         /* device version */
	DWORD  cnt;				/* loop counter */

	/*
	 * verify USBDMX dll version - this generates a error if this code was 
	 * compiled using a more recent usbdmx.h that the dll found. 
	 */
	if (!USBDMX_DLL_VERSION_CHECK())
	{
		printf("USBDMX.DLL version does not match, giving up!\n");
		printf("found 0x%x, expected 0x%x\n", usbdmx_version(), USBDMX_DLL_VERSION);
		getchar();
		return 1;
	}
	/*
	 * special handling for RDM: version 0x1000 is know to not fully respect
	 * the RDM transmitter timings -> print a warning.
	 */
	if (usbdmx_version() == 0x1000)
	{
		printf("USBDMX.DLL version 0x%x is known to not fully comply with the RDM transmitter timings.\n", usbdmx_version());
		printf("please consider upgrading.\n");
	}

	printf("Using USBDMX.DLL version 0x%x\n\n", usbdmx_version());

	/* open the first devices 
	 * the for() loop is not needed, but recommended as all Peperoni devices
	 * are in a single list. Every newly connected device is added to the 
	 * next available place in the list. List entries are marked as empty
	 * if the device as been disconnected AND all handles to this device
	 * have been closed. So under special conditions list entries can be empty
	 * followed by non-empty places. */
	h = INVALID_HANDLE_VALUE;	/* avoid "potentially uninitialized variable" warnings */
	ret = FALSE;				/* avoid "potentially uninitialized variable" warnings */
	for (device = 0; device < 16; ++device)
	{
		ret = usbdmx_open(device, &h);
		if (ret == TRUE)
			break;
	}
	if (ret == FALSE)
	{
		printf("no usbdmx-interface available, giving up!\n");
		return 1;
	}

	/* identify the interface */
	if (!usbdmx_product_get(h, product, sizeof(product)))
		printf("ERROR: reading product string failed.\n");
	if (!usbdmx_device_version(h, &version))
		printf("ERROR: reading version failed.\n");

	/* print product string and version */
	printf("Found a %ws with firmware version 0x%04lx.\n\n", product, version);

	/* run rdm discovery */
	if (discovery(h) < 1)
	{
		printf("no devices found\n");
	}
	else
	{
		printf("found %u devices.\n", TodDevices);

		/* loop over all devices we discovered */
		for (cnt = 0; cnt < TodDevices; ++cnt)
		{
			testDevice(h, &ToD[cnt]);
		}
	}

	/* close handles to interfaces */
	usbdmx_close(h);

	printf("Program finished\n");
	getchar();

	return 0;
}
